/*
                                  fft.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*       Part of:        SkyMaker
*
*       Author:         E.BERTIN (IAP)
*
*       Contents:       Routines dealing with FFT.
*
*       Last modify:    10/10/2000 (modified to make correlation)
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "define.h"
#include "fft.h"

#define	SWAP(a,b)	tempr=(a);(a)=(b);(b)=tempr

/********************************** fourn ************************************/
/*
ndim-dimensional Fast Fourier Transform (from Numerical Recipes in C, p.468).
The 1-dimensional 'data' array is replaced by its Fourier transform:
storage in 'data': (Re[0][0],Im[0][0]); (Re[0][1],Im[0][1]);...
The 1-dimensional 'nn' array contains the sizes in each dimension.
isign = +1 for direct FFT, or -1 for inverse FFT.
Note: float version.
*/
void fourn(PIXTYPE *data, unsigned int *nn, int ndim, int isign)
  {
   int		idim;
   unsigned int	i1,i2,i3,i2rev,ip1,ip2,ip3,ifp1,ifp2;
   unsigned int	ibit,n,nprev,nrem,ntot;
   PIXTYPE	tempi,tempr, *d3, *d3rev, *dk1,*dk2;
   float	theta,wi,wpi,wpr,wr,wtemp;

/* a small trick to begin arrays with nb 1 (a reminiscence of FORTRAN!) */
  data--;

  ntot=1;
  for (idim=0;idim<ndim;idim++)
    ntot *= nn[idim];
  nprev=1;
  for (idim=ndim;idim--;)
    {
    n=nn[idim];
    nrem=ntot/(n*nprev);
    ip1=nprev << 1;
    ip2=ip1*n;
    ip3=ip2*nrem;
    i2rev=1;
    for (i2=1;i2<=ip2;i2+=ip1)
      {
      if (i2 < i2rev)
        {
        for (i1=i2;i1<=i2+ip1-2;i1+=2)
          {
          for (i3=i1;i3<=ip3;i3+=ip2)
            {
            d3rev=i2rev+(d3=data+i3)-i2;
            SWAP(*d3,*d3rev);
            SWAP(*(d3+1),*(d3rev+1));
            }
          }
        }
      ibit=ip2 >> 1;
      while (ibit >= ip1 && i2rev > ibit)
        {
        i2rev -= ibit;
        ibit >>= 1;
        }
      i2rev += ibit;
      }
    ifp1=ip1;
    while (ifp1 < ip2)
      {
      ifp2=ifp1 << 1;
      theta=isign*6.28318530717959/(ifp2/ip1);
      wtemp=sin(0.5*theta);
      wpr = -2.0*wtemp*wtemp;
      wpi=sin(theta);
      wr=1.0;
      wi=0.0;
      for (i3=1;i3<=ifp1;i3+=ip1)
        {
        for (i1=i3;i1<=i3+ip1-2;i1+=2)
          {
          for (i2=i1;i2<=ip3;i2+=ifp2)
            {
            dk2=(dk1=data+i2)+ifp1;
            tempr=wr**dk2 - wi**(dk2+1);
            tempi=wr**(dk2+1) + wi**dk2;
            *(dk2++)=*dk1-tempr;
            *dk2=*(++dk1)-tempi;
            *(dk1--) += tempi;
            *dk1 += tempr;
            }
          }
        wr+=(wtemp=wr)*wpr-wi*wpi;
        wi+=wi*wpr+wtemp*wpi;
        }
      ifp1=ifp2;
      }
    nprev *= n;
    }

  return;
  }



/********************************* fastconv **********************************/
/*
Optimized 2-dimensional FFT convolution, dedicated to 2D real data (images). 
Returns the Nyquist vertical frequency vector (for which memory is allocated).
*/
PIXTYPE *fastconv(PIXTYPE *image, PIXTYPE *mask, PIXTYPE *speqmask,
	int width, int height)
  {
   int		i, nbfreq;
   PIXTYPE	*speqimage, *imap, *maskp,
		real, imag, fac;

/* Allocate temporary memory for storing Nyquist unfolded freq. components */
  QMALLOC(speqimage, PIXTYPE, 2*height);
  if (!speqmask)
    {
    QMALLOC(speqmask, PIXTYPE, 2*height);
/*-- Forward FFT for mask */
    frt2d(mask, speqmask, width, height, 1);
    }
/* Forward FFT for image */
  frt2d(image, speqimage, width, height, 1);
/* Actual convolution (Fourier product) */
  nbfreq = width*height/2.0;
  fac = 1.0/nbfreq;
  imap = image;
  maskp = mask;
  for (i=0; i<nbfreq; i++, maskp+=2)
    {
    real = *imap**maskp-*(imap+1)**(maskp+1);
    imag = *imap**(maskp+1)+*(imap+1)**maskp;
    *(imap++) = fac*real;
    *(imap++) = fac*imag;
    }

  nbfreq = height;
  imap = speqimage;
  maskp = speqmask;
  for (i=0; i<nbfreq; i++, maskp+=2)
    {
    real = *imap**maskp-*(imap+1)**(maskp+1);
    imag = *imap**(maskp+1)+*(imap+1)**maskp;
    *(imap++) = fac*real;
    *(imap++) = fac*imag;
    }

/* Reverse FFT */
  frt2d(image, speqimage, width, height, -1);
  free(speqimage);
  
return speqmask;
  }




/********************************* fastcorr **********************************/
/*
Optimized 2-dimensional FFT correlation, dedicated to 2D real data (images). 
Returns the Nyquist vertical frequency vector (for which memory is allocated).
*/
void fastcorr(PIXTYPE *image, PIXTYPE *ref,
	int width, int height, PIXTYPE *kern, int kw, int kh, int *xmax, int *ymax, float *nsigma)
  {
   int		i,nbfreq,j, ntot;
   PIXTYPE	*speqref, *speqimage, *speqmask, *imap, *refp, *maskp,
		real, imag, fac;
   float        mean, sigma, diff, vmax;

 /*  float *b, *tb, *timage;    */

  float *mask, *npmask, *ppmask, *npkern, *ppkern;

/* Allocate temporary memory for storing Nyquist unfolded freq. components */

  QMALLOC(speqimage, PIXTYPE, 2*height);

  QMALLOC(speqref, PIXTYPE, 2*height);

if (kw >0 && kh >0) {

/* Prepare kernel */
 mask= (float *) malloc(width*height*sizeof(float));
  QMALLOC(speqmask, PIXTYPE, 2*height);

ppmask=mask;
for (i=0;i<width*height;i++) {*(ppmask++)=0.00;}

for (j=0;j<kh;j++)  {

  npkern=ppkern=(kern+j*kw);

 if (j<(kh-1)/2) {        /* negative end of kernel => upper part of mask */
  ppmask=npmask=(mask+(height-(kh-1)/2+j)*width);
  /*printf ("N kernel row %d -> mask %d \n", j,(height-(kh-1)/2+j) );*/
 } else {
  ppmask=npmask=(mask+(j-(kh-1)/2)*width);
  /*printf ("P kernel row %d -> mask %d \n", j,(j-(kh-1)/2) );*/
}

  npmask += (width-(kw-1)/2);  /* start to store from here */
  ppkern += (kw-1)/2;      /* The positive part starts here */

  for (i=0;i<kw; i++) {
     if (i<(kw-1)/2) {
       *(npmask++)= *(npkern++);
     }  else {
        *(ppmask++)= *(ppkern++);
     }
   }

 }

}


/*printf ("Forward start %d %d\n",width, height);*/


/*-- Forward FFT for reference image */
     frt2d(ref, speqref, width, height, 1); 
    
/* Forward FFT for image */
  frt2d(image, speqimage, width, height, 1);

  /*printf ("Doing correlation\n");*/

/* Actual correlation (Fourier product) */
  nbfreq = width*height/2.0;
  fac = 1.0/nbfreq;
  imap = image;
  refp = ref;

  for (i=0; i<nbfreq; i++, refp+=2)
    {
 /* Sign changed wrt convolution */     
    real = *imap**refp+*(imap+1)**(refp+1);
    imag = *imap**(refp+1)-*(imap+1)**refp;
    *(imap++) = fac*real;
    *(imap++) = fac*imag;

    }

  nbfreq = height;
  imap = speqimage;
  refp = speqref;

  for (i=0; i<nbfreq; i++, refp+=2)
    {
/* Sign changed wrt convolution */
    real = *imap**refp+*(imap+1)**(refp+1);
    imag = *imap**(refp+1)-*(imap+1)**refp;
    *(imap++) = fac*real;
    *(imap++) = fac*imag;
    }



if (kw >0 && kh >0) {

/* FORWARD FFT for kernel */

 frt2d(mask,speqmask, width, height, 1);


/* Convolve with given kernel    */

 /*printf("Convolving with kernel\n");*/

/* Actual convolution (Fourier product) */
  nbfreq = width*height/2.0;
  fac = 1.0/nbfreq;
fac=1.0;
  imap = image;
  maskp = mask;

  for (i=0; i<nbfreq; i++, maskp+=2)
    {

    real = *imap**maskp-*(imap+1)**(maskp+1);
    imag = *imap**(maskp+1)+*(imap+1)**maskp;
    *(imap++) = fac*real;
    *(imap++) = fac*imag;

    }

  nbfreq = height;
  imap = speqimage;
  maskp = speqmask;

  for (i=0; i<nbfreq; i++, maskp+=2)
    {
/* Sign changed wrt convolution */
    real = *imap**maskp-*(imap+1)**(maskp+1);
    imag = *imap**(maskp+1)+*(imap+1)**maskp;
    *(imap++) = fac*real;
    *(imap++) = fac*imag;
    }

   free(mask);
   free(speqmask);

 }


/*printf ("Reversing FFT\n");*/

/* Reverse FFT */
 frt2d(image, speqimage, width, height, -1);

 /*printf ("done\n");*/


/*Compute maximum and sigma */
imap=image;
vmax=0.;
*xmax=-1;
*ymax=-1;
mean=0.;
ntot=0;

for (i=0;i<height;i++) {
 for (j=0;j<width;j++) {
    if (*imap > vmax) {
        vmax=*imap;
        *xmax=j;
        *ymax=i;
    }
  if (*imap>0) {
        mean+=*imap;
        ntot++;
    }
   imap++;
 }
}


/*check if wrap around  */
 if (*xmax >width/2. || *ymax > height/2.) {
   if (*xmax > width/2.) {
          *xmax-=width;
    }
   if (*ymax > height/2.) {
         *ymax-=height;
    }
 }



mean/=ntot;

sigma=0.;
imap=image;

for (i=0;i<(width*height);i++) {
  if (*imap>0) {
      diff=*imap-mean;
      sigma+=(diff*diff);
   }
   imap++;
}

sigma=sqrt(sigma/ntot);

*nsigma=vmax/sigma;

/*printf ("Maximum %.2f at %d %d, sigma (%d: %.2f): %.2f\n", vmax, *xmax, *ymax, ntot, sigma, *nsigma);*/




/* Free memory */

  free (speqimage);
  free (speqref);

  /*printf ("freed\n");*/

  }

/*********************************** frt2d ***********************************/
/*
Fast real 2D Fourier transform, based on FFT (inspired from Numerical Recipes
in C, 2nd ed., p.528). 
*/
void frt2d(PIXTYPE *data, PIXTYPE *speq, int width, int height, int isign)
  {
   unsigned int	i2,i3,ii3,j2,j3, nn[2], uwidth, uheight;
   PIXTYPE	*datai2,*datai21, *dataj2,*dataj21, *speqj2,*speqj21,
		c2,h1r,h1i,h2r,h2i;
   float	theta, wi,wpi,wr,wpr,wtemp;

  uwidth = width;
  uheight = height;
  theta = isign*6.28318530717959/height;
  wtemp = sin(0.5*theta);
  wpr = -2.0*wtemp*wtemp;
  wpi = sin(theta);
  nn[0] = uheight;
  nn[1] = uwidth>>1;
  c2 = -0.5*isign;
  if (isign==1)
    {
    fourn(data, nn, 2, isign);
    datai2 = data;
    speqj2 = speq;
    for (i2=0; i2<uheight; i2++, datai2 += uwidth)
      {
      *(speqj2++) = *(datai2++);
      *(speqj2++) = *(datai2--);
      }
    }
  wr = 1.0;
  wi = 0.0;
  for (i3=0,ii3=0; i3<=(uwidth>>2); i3++,ii3+=2)
    {
    if (!i3)
      {
      datai2 = data;
      for (i2=0; i2<uheight; i2++, datai2+=uwidth)
        {
        datai21 = datai2 + 1;
        j2 = i2? ((uheight-i2-1)<<1)+2 : 0;
        speqj2 = speq+j2;
        speqj21 = speqj2+1;
        h1r = 0.5*(*datai2+*speqj2);
        h1i = 0.5*(*datai21-*speqj21);
        h2i = c2*(*datai2-*speqj2);
        h2r = -c2*(*datai21+*speqj21);
        *datai2 = h1r+h2r;
        *datai21 = h1i+h2i;
        *speqj2 = h1r-h2r;
        *speqj21 = h2i-h1i;
        }
      }
    else
      {
      datai2 = data+ii3;
      for (i2=0; i2<uheight; i2++, datai2+=uwidth)
        {
        datai21 = datai2 + 1;
        j2 = i2? uheight-i2 : 0;
        j3 = uwidth+2-((i3+1)<<1);
        dataj2 = data + j2*uwidth + j3;
        dataj21 = dataj2 + 1;
        h1r = 0.5*(*datai2+*dataj2);
        h1i = 0.5*(*datai21-*dataj21);
        h2i = c2*(*datai2-*dataj2);
        h2r = -c2*(*datai21+*dataj21);
        *datai2 = h1r+wr*h2r-wi*h2i;
        *datai21 = h1i+wr*h2i+wi*h2r;
        *dataj2 = h1r-wr*h2r+wi*h2i;
        *dataj21 = -h1i+wr*h2i+wi*h2r;
	
        }
      }
    wr = (wtemp=wr)*wpr-wi*wpi+wr;
    wi = wi*wpr+wtemp*wpi+wi;
    }

  if (isign == -1)
    fourn(data, nn, 2, isign);


  return;
  }

